
(*
val _ = app load ["Types", "Real", "Int", "Mosml"] ;
fun uc () = use "count.sml" ;
*)
structure Count = struct 
local open Types Gen ;
in

(* getpref : (int -> bool) -> substring -> int * substring
  f indicates whether candidate is continuing, 
  find first candidate in arg2 satisfying f, return it and rest of arg2 *)
fun getpref f ss = 
  let open Substring ;
    val dr = dropl (not o f o ord) ss ;
  in case getc dr of NONE => (0, dr) 
      | SOME (c, dr') => (ord c, dr') end ;

(* allocv : (int -> bool) -> group list -> int -> int * (int * vote list) list
  count votes for a candidate no longer continuing;
  the votes may be in several groups, since they may have gone to the candidate
  at several counts;
  divide votes according to next preference, return list of one list
  of votes per candidate, each with number of votes; also votes with no
  further preference treated as candidate 0;
  also returns total tvnum of votes counted *) 
fun allocv f vss nc = 
  let val a = Array.array (nc + 1, (0, []))
    (* count a single vote *)
    fun count1 (Vote (cp)) = 
      let val (p, ps) = getpref f cp ;
      in updatef a p (fn (nv, vs) => (nv + 1, Vote (ps) :: vs)) end ;
    (* count the list of vote in a single group *)
    fun count1list (Group {votes, tvnum, ...} : group) = (
      app count1 votes ; tvnum) ;
    (* count a list of groups of votes *)
    val totval = List.foldl (op +) 0 (map count1list vss) ;
  in (totval, toList a) end ;

(* countE' : (int -> bool) -> int -> 
  (int * vote list) list -> tv -> bool -> group list 
  turn result of counting votes into groups;
  applies new tv given as argument;
  omits groups for non-continuing candidates;
  returns list of new groups;
  includes group for candidate 0 (exhausted votes) but
  gives them zero value in case of a surplus *)

fun countE' f cnt al (tv as TV (numer, denom)) isSurp = 
  let 
    (* mkng : cand -> int * vote list -> group
      make new group of votes for one candidate from result of count *)
    (* in case of a surplus, votes transferred to candidate 0
      (ie, exhausted) have no value, but they do have a value when
      a candidate has been excluded *)
    fun mkng i (nv, vs) = Group {cand = i, tv = tv, votes = vs, num = nv,
	count = cnt, tvnum = 
	  case (i, isSurp) of (0, true) => 0 
	    | _ => (nv * numer) div denom} 
      handle exn => raise Fail ("countE', mkng: " ^ exnMessage exn) ;
    (* mkngs : cand -> (int * vote list) list -> group
      make new group of votes for each candidate starting from i,
	but exclude non-continuing candidates *)
    fun mkngs i [] = []
      | mkngs i ((x as (nv, _)) :: xs) =
	(* include empty groups at this stage *)
	if i = 0 orelse f i then mkng i x :: mkngs (i+1) xs 
	else if nv = 0 then mkngs (i+1) xs
	else raise Fail "mkngs : nv <> 0, f i <> true" ;
    val newGroups = mkngs 0 al ;
  in newGroups end ; 

(* checking - only for MosML interactive (or use PolyML.print for printVal) *)
(*
fun checkLoss groups newGroups = 
  let fun ntvn (Group {num, tvnum, cand, ...}) = (cand, num, tvnum) ;
  in if tvof groups = TV (1, 1) then () else (
      printVal (cnt, tvof groups, map ntvn groups, map ntvn newGroups) ;
      print "\n") end ;
*)

(* countE : (int -> bool) -> int -> group list -> int -> int * group list
  count list of groups of votes, upon exclusion of candidate, at count cnt *)
fun countE f cnt groups nc = 
  let val (totval, al) = allocv f groups nc ;
    val newGroups = countE' f cnt al (tvof groups) false ;
  in (totval, newGroups) end ;

(* futv : tv -> group -> group list list -> group list list 
  where arg3 is list of lists of vote groups, grouped according to tv
  in descending order, insert group (with tv) in appropriate list *)
fun futv tv group [] = [[group]]
  | futv tv group (gs :: gss) =
    case tvcompare (tv, tvof gs) of LESS => gs :: futv tv group gss
      | EQUAL => (group :: gs) :: gss
      | GREATER => [group] :: gs :: gss ;

(* ins : int * 'a -> (int * 'a) list -> (int * 'a) list
  insert pair into list which is ordered by first component *)
fun ins (i, x) [] = [(i, x)]
  | ins (i, x) (yys as (j, y) :: ys) = 
    if i < j then (i, x) :: yys
    else if i > j then (j, y) :: ins (i, x) ys
    else raise Fail "ins - equal first components" ;

(* tot' : ('a * int) list -> int, get total of second components *) 
fun tot' [] = 0 | tot' ((i, x) :: ixs) = x + tot' ixs ;
fun tot x = (tot' x, x) ;

(* filcce : info -> cand list -> (cand * 'b) list -> (cand * 'b) list
  arg 2 is list of some candidates (in order);
  filcce filters arg 3 (list of pairs) to include those where the
  candidate is in arg 2 or has non-zero vote total *)
fun filcce (Info {candTot, ...}) ccf cxsf =
  let 
    fun f [] [] = []
      | f [] cxs = 
	List.filter (fn (cand, x) => Array.sub (candTot, cand) <> 0) cxs
      | f (cc as c :: cc') ((cand, x) :: cxs) = 
	if cand = c then (cand, x) :: f cc' cxs
	else if Array.sub (candTot, cand) <> 0 then (cand, x) :: f cc cxs
	else f cc cxs 
      | f _ _ = raise Fail "Match in filcce" ;
  in f ccf cxsf end ;

(* elect : info -> cand -> unit
  upon election of candidate, update candStatus and nelec *)
fun elect (Info {candStatus, nelec, ...}) cand =
  let 
    (* mkElND : change candidate status for candidate elected *)
    fun mkElND (Cont x) = ElND x
      | mkElND _ = raise Fail "mkElND" ;
  in nelec := !nelec + 1 ; updatef candStatus cand mkElND end ;

(* upd1 : info -> group -> elec option
  for one group of newly counted votes, update relevant candidate,
  but ignore empty groups;
  see if new candidate elected, if so update nelec and candStatus,
  and return SOME (Elec ...) *)
fun upd1 info (group as Group {num = 0, ...}) = NONE
  | upd1 (info as 
	Info {candStatus, candTot, count, fixed = {nq, ...}, nelec, ...})
      (group as Group {cand, tvnum, tv, ...} : group) = (
    updatef candTot cand (fn n => n + tvnum) ;
    (* for a real candidate, add group, check if elected *)
    if cand <> 0 then (
      updatef candStatus cand 
	(fn Cont v => Cont (futv tv group v)
	  | _ => raise Fail ("updating status for status not Cont, cand " ^
	    Int.toString cand)) ;
      (* check candidate's new count to see if above quota *)
      case Array.sub (candTot, cand) of n =>
	if n >= nq then 
	  (elect info cand ; SOME (Elec (cand, n, group, !count)))
	else NONE)
    else NONE) ; 

(* updateCounts : info -> (cand * int) option * group list -> elec list
  given a list of newly counted groups, update the candidate status
  (total count and list of groups);
  argument cntvo is NONE for first count,
  then candidate whose votes have been transferred, and change
  to that candidate's total (a negative value);
  return, and update events with, newly elected candidates *)
fun updateCounts (info as Info {candStatus, candTot, count, events, results, 
    fixed = {nq, ne, ...}, nelec, ...}) (cntvo, newGroups) = 
  let 
    (* collect results of the count as weighted votes *)
    fun f (Group {tvnum, cand, ...}) = (cand, tvnum) ;
    (* include votes just counted, negatively, for candidate
      from whom they were transferred *)
    val newres = case cntvo of NONE => (map f newGroups) 
      | SOME cntv => ins cntv (map f newGroups) ;
    (* find candidates whose total is recalculated (even if unchanged) *)
    val cc as 0 :: _ = map #1 newres ;

    (* collect results of the count as numbers of votes *)
    fun g (Group {num, cand, ...}) = (cand, num) ;
    val countres = map g newGroups ;

    (* update info for all new groups of votes,
      and get list of newly elected candidates *)
    val newCandopts = map (upd1 info) newGroups ;
    val newCand3s' = List.mapPartial (fn x => x) newCandopts ;

    (* get newly elected candidates in the right order *)
    val newCand3s = Tie.sortElect newCand3s' info ;
    val () = recordEvent info (Elect (map Tie.ue1 newCand3s)) ;

    (* update total for candidate whose group(s) were distributed,
      this was not done by upd1 *)
    val () = case cntvo of NONE => ()
      | SOME (dc, ntv) => updatef candTot dc (fn tot => tot + ntv) ; 

    (* get list of new candidate totals *)
    val totres = filcce info cc (numl 0 (toList candTot)) ;
    
    val () = push (!count, Cres {tot = tot totres, tv = tvof newGroups,
      new = tot newres, count = tot countres}) results ;
    val () = if !nelec >= ne then raise Finished else () ; (* cl 4(1) *) 
  in newCand3s end ;

(* csf : status array -> cand -> bool, tell if candidate is continuing *)
fun csf _ 0 = raise Fail "program error : asking status of candidate 0"
  | csf candStatus i = 
    case (Array.sub (candStatus, i)) of Cont _ => true
      | _ => false ;

(* distGroups : info -> cand -> group list -> elec list
  distribute a given list of groups of votes (of a single tv)
  of an excluded candidate 
  to the next candidates, and update counts accordingly *)
fun distGroups (info as Info 
      {candStatus, fixed = {nc, ...}, count, events, ...}) cand groups = 
  let
    val () = count := !count + 1 ;
    val countDetails as (exclcand, tv, cts) = 
      (cOfG (hd groups), tvof groups, map ctOfG groups) ;
    val true = cand = exclcand ;
    val () = recordEvent info (ExclDist countDetails) ;
    val (totval, ngs) = countE (csf candStatus) (!count) groups nc ;
    val newCands = updateCounts info (SOME (exclcand, ~totval), ngs) ; 
  in newCands end ;

(* exclCand : info -> cand -> elec list
  exclude a candidate, and return lists of elected candidates;
  will cover several counts, one for each vote value *)
fun exclCand (info as Info {candStatus, ncont, fixed = {ne, ...}, ...}) cand = 
  let
    val () = recordEvent info (ExclDet cand) ;
    val () = ncont := !ncont - 1 ;
    val bs = case Array.sub (candStatus, cand) of Cont bs' => bs'
        | _ => raise Fail "exclCand : status already not Cont" ; 
    val () = Array.update (candStatus, cand, Excluded) ; 
    val ncss = (map (distGroups info cand) bs) ;
  in List.concat ncss end ;

(* getcss : status array -> int array -> (cand * int) list
  get list of counts of candidates currently in contest *)
fun getcss candStatus candTot = 
  let 
    (* note - we exclude candidate 0 here *)
    val (_ :: cslist) = toList candStatus ;
    fun mkcns i [] = []
      | mkcns i (Cont _ :: css) = 
        let val n = Array.sub (candTot, i) in (i, n) :: mkcns (i+1) css end
      | mkcns i (_ :: css) = mkcns (i+1) css ;
  in mkcns 1 cslist end handle _ => raise Fail "getcss" ;

(* foldr1 : ('a -> 'b -> 'b) -> ('a -> 'b) -> 'a list -> 'b
  sort of generalized fold, for non-empty list *)
fun foldr1 f g [] = raise Fail "foldr1 []"
  | foldr1 f g [x] = g x
  | foldr1 f g (x :: xs) = f x (foldr1 f g xs) ;

(* lowestCand' : info -> cand list * int
  gets list of equal lowest candidates *) 
fun lowestCand' (info as Info {candStatus, candTot, ...} : info) = 
  let val css = getcss candStatus candTot ;
    (* here we get list of candidates of equal lowest count *)
    fun f (i, n) (il, n') = if Int.< (n, n') then ([i], n) 
      else if n = n' then (i :: il, n)
      else (il, n') ;
    fun g (i, n) = ([i], n) ;
  in foldr1 f g css end ;

(* lowestCand : info -> cand, gets lowest candidate *) 
fun lowestCand (info as Info {count, events,  ...}) = 
  case lowestCand' info of ([low], _) => low
    | (lows, n) => 
      let val (cnt, c) = Tie.chooseExcl lows info ;
        val () = recordEvent info (ExclTie 
	    {lows = lows, count = cnt, excl = c, votes = n}) ;
      in c end ;

(* findgp : info -> elec -> group
  find group of votes allocated to successful candidate,
  whose surplus votes have not yet been distributed (status is ElND _),
  at the count when he gained a quota *)
fun findgp (Info {candStatus, ...} : info) (Elec (cand, _, group, cnt)) = 
  let val ElND gpss = Array.sub (candStatus, cand) ;
    val gphs = map hd gpss ;
    fun p (Group {count, ...}) = count = cnt ;
    (*
    val true = [group] = List.filter p gphs ;
    *)
    val true = (cOfG group, ctOfG group) = (cand, cnt) ;
  in group end 
  handle _ => raise Fail ("findgp : cand " ^ Int.toString cand ^
      ", count " ^ Int.toString cnt) ;

val eref = ref (~1, ~1, ~1) ;

(* distElec : info -> elec -> elec list 
  distSurp : info -> elec -> elec list 
  distribute surplus votes for elected candidate,
  changes his/her status from ElND _ to Elected,
  return list of candidates successful as a consequence;
  note - distSurp does the counting of the votes to distribute the surplus,
  which is not done when the candidate achieves the quota exactly;
  the remainder of distElec is done for all elected candidates 
  *)
fun distSurp (info as Info {count, events, fixed as {nq, nc, ...}, 
    candStatus, ...}) 
    (elec as Elec (cand, n, 
      group as Group {tv as TV (tvn, tvd), num, ...}, cnt)) = 
  let
    val () = count := !count + 1 ;
    (* count votes to be distributed *)
    val (_, al) = allocv (csf candStatus) [group] nc ;

    (* calculate new transfer value *)
    val surp = n - nq ; (* candidate's surplus votes (weighted) *)
    val nexh = (#1 (hd al)) ; (* number of counted votes exhausted (actual) *)
    val nnp = num - nexh ; (* number of votes to be transferred (actual) *)
    (* omit this check - actually OK provided num is not also zero 
    val () = if nnp = 0 then raise Fail "distSurp: nnp = 0" else () ;
    *)
    val newtv' = TV (surp, nnp) ;
    (* see if clause 1(2) applies, and, if so, report number lost *)
    val newtv = 
      (case tvcompare (newtv', tv) of GREATER =>
	  let val nlost = (surp - (tvn * nnp) div tvd) ;
	  in (recordEvent info (Cl_1_2 nlost) ; tv) end
	| _ => newtv')
      handle exn => raise Fail ("distSurp, newtv: " ^ exnMessage exn) ;

    val () = 
      case newtv of (TV (0, 0)) => raise Fail "distSurp: newtv 0/0" 
        | (TV (_, 0)) => raise Fail "distSurp: newtv denom = 0" 
	| _ => () ;
    val () = recordEvent info (Surplus (cand, newtv, cnt)) ;
    val newGroups = countE' (csf candStatus) (!count) al newtv true ;
    val newCands = updateCounts info (SOME (cand, ~surp), newGroups) ; 
  in newCands end ;

fun distElec (info as Info {count, events, fixed as {nq, nc, ...}, 
    candStatus, ...}) 
    (elec as Elec (cand, n, 
      group as Group {tv as TV (tvn, tvd), num, ...}, cnt)) = 
  let
    (* change candidate status from ElND _ to Elected *)
    fun mkel (ElND _) = Elected
      | mkel _ = raise Fail "distElec : mkel" ;
    val () = updatef candStatus cand mkel ;
  in if n - nq = 0 then (* no surplus *)
      (recordEvent info (SurpZero {cand = cand, count = cnt}) ; [])
    else distSurp info elec end ;
  
(* findConts : status array -> int list
  find continuing candidates, check no ElND candidates *)
fun findConts candStatus = 
  let fun f (i, Cont _, clist) = i :: clist
      | f (_, ElND _, _) = raise Fail "findConts"
      | f (_, _, clist) = clist ;
  in Array.foldri f [] (candStatus, 1, NONE) end ;

(* prefCount' : elec list * info -> info
  complete counting after initial distribution, where 
  arg elected is those elected on initial count *)
fun prefCount' (elected, 
    info as Info {ncont, nelec, fixed = {ne, ...}, candStatus, ...}) =
  let
    (* next : elec list -> unit
      if candidates elected, distribute surplus, and recursively;
      if none, exclude lowest *)
    fun next [] = 
	(* test if finished under cl 4(2) - at this point, any previously
	  excluded candidate has had votes dealt with, and no elected
	  candidate with surplus not yet distributed *)
	if !ncont > ne then next (exclCand info (lowestCand info))
	else (* only right number of candidates remaining *)
	  let val conts = findConts candStatus ;
	    val _ = if length conts <> !ncont - !nelec then 
		raise Fail "prefCount': length of conts"
	      else if conts = [] then 
		raise Fail "prefCount': conts is []"
	      else () ;
	  in recordEvent info (FinCont conts) end 
      | next elected = 
	let val newes = map (distElec info) elected ;
	in next (List.concat newes) end ;

  in next elected ; info end
  handle Finished => info
    | exn => (inforef := info ; raise exn) ;

(* prepCount : int * int * int * vote list -> elec list * info
  set up info structure, do count of first preferences *) 
fun prepCount (nc, ne, nv, deceased, commLot, votes) =
  let
    val initGroup = Group {cand = 0, tv = TV (1, 1), num = nv, tvnum = nv, 
      count = 0, votes = map (fn v => Vote (Substring.all v)) votes}  ;
    val nq = nv div (ne + 1) + 1 ; (* quota *)
    val info : info = Info
      {candStatus = Array.array (nc + 1, Cont []), 
        candTot = Array.array (nc + 1, 0), 
        ncont = ref (nc - length deceased), nelec = ref 0, 
	fixed = {nc = nc, ne = ne, nq = nq, nv = nv}, 
        count = ref 1, results = ref [], events = ref [(1, InitDecd deceased)],
	commLot = commLot} ;

    (* TODO - commLot function that uses user input *)

    (* update candStatus for deceased candidates *)
    val Info {candStatus, ...} = info ;
    fun dodec dc = Array.update (candStatus, dc, Deceased) ;
    val () = app dodec deceased ;

    (* perform initial count; if no deceased candidates, don't
      bother looking up status of candidates *)
    val (_, initNewGroups) = 
      case deceased of [] => countE (fn _ => true) 1 [initGroup] nc 
	| _ => countE (csf candStatus) 1 [initGroup] nc ;

    val elected = updateCounts info (NONE, initNewGroups) ; 
  in (elected, info) end ;

(* prefCount : int * int * int * vote list -> info
  count votes according to Hare-Clark system *)
fun prefCount args = prefCount' (prepCount args) ;

end ; (* local open Types ; *)
end ; (* structure Count *)

(* testing *)
(*
val nc = 20 ; val nv = 20000 ; val ne = 5 ;
val votes = Ran.readsvotes "votes20" nv ;

(*
load "Ran" ;
*)

val votes = Ran.ranpvotes nc nv ;

val nc = 49 ; val nv as nvr = 90000 ; val ne = 7 ;
val votes = Ran.readsvotes "votes49" nvr ;
val nc = 49 ; val nv = 50000 ; val ne = 7 ;
val votes = List.drop (votes, nvr - nv) ;
val nc = 49 ; val nv = 120000 ; val ne = 7 ;
val votes = votes @ votes ;

open Count ;
fun t () = 
  let val r = prefCount (nc, ne, nv, votes) ; in fn () => r end ;

val info as Info {candStatus : status array, candTot : int array,
  count : int ref, events : (int * elecEvent) list ref,
  fixed : {nc : int, ne : int, nq : int}, ncont : int ref, nelec,
  results : (int * results) list ref} = t () () ;
print "\a" ;

fun r i = List.filter (fn (c, _) => c = i) (!results) ;
fun e i = List.filter (fn (c, _) => c = i) (!events) ;
val csl = List.filter (fn (_, str) => str <> "Excluded")
  (numl 0 (map statusString (Array.foldr (op ::) [] candStatus))) ;

*)

(* call tree

WHEN CANDIDATE EXCLUDED 

prefCount'

  exclCand (called once for each candidate to be excluded)

    distGroups (called once for each resulting count, 
	as there is one count for each vote value,
	and the sequence of them is not interrupted by candidates 
	becoming successful)

      countE 
      updateCounts

	allocv
	countE'

-----------------

DEAL WITH SUCCESSFUL CANDIDATE 
(surplus, if any, distributed) 

prefCount'

  distElec
    distSurp (not if candidate achieves quota exactly)

      allocv 
      countE' 
      updateCounts

---------

INITIAL COUNT 

prefCount'

  prepCount 

    countE 
    updateCounts

NOTES :
count is incremented at the start of distGroups and of distSurp

The sequence of counts is determined within the function next,
which is internal to prefCount'.
The argument to next is a list of successful candidates whose 
surpluses need distributing.

The functions exclCand, distGroups, distSurp and updateCounts
return the newly elected candidates.
*)
