
(*
load "Msp" ;
load "ListPair" ;
load "TextIO" ;
*)
structure Tab = struct 

local open Msp ; infix && ; open Types ;
in

fun i2s i = if i < 0 then "-" ^ Int.toString (~i) else Int.toString i ;

(* wsc : wseq list -> wseq, concatenate list of wseqs *)
fun wsc [] = Empty
  | wsc (ws :: wss) = ws && wsc wss ;

(* row : wseq list -> wseq 
  srow, shead : string -> string list -> wseq
  srowc : string -> wseq list -> wseq
  tab : wseq list * wseq list list -> wseq 
  create HTML table, row *)
fun srow opt strs = tra opt (wsc (List.map (td o $) strs) && Nl) ;
fun srowc opt cells = tra opt (wsc cells && Nl) ;
fun shead opt strs = tha opt (wsc (List.map (td o $) strs) && Nl) ;
fun row wss = tr (wsc (List.map td wss) && Nl) ;
fun head wss = th (wsc (List.map td wss) && Nl) ;

fun tab (h, rs) = tablea "border" (wsc (head h :: (List.map row rs))) ;

(* getfds : (resrec -> 'a) -> info -> (int * 'a) list
  get part of each results entry *)
fun getfd sel (cnt, Cres r) = (cnt, sel r) ; 
fun getfds sel (Info {results = ref res, ...}) = Gen.maprev (getfd sel) res [] ;
fun revres (Info {results = ref res, ...}) = rev res ;
 
(* scrheads : int -> string list, make headers of scrutiny sheet *)
fun scrheads nc = "Cnt" :: 
  List.map i2s (Gen.upto (1, nc)) @ 
  ["Exh", "Total", "TV"] ;

(* mkrow', mkrow : int -> (cand * int) list -> string list 
  make entries relating to candidates 1 to nc in a row of scrutiny sheet *)
fun mkrow' nc cns = 
  let 
    (* mkr : cand -> (cand * int) list, 
      get strings for numbers for candidates c to nc, blank string
      where candidate missing from arg2 *)
    fun mkr c (cns as (c', n) :: cns') = 
	if c' > c then "" :: mkr (c+1) cns 
	else if c = c' then i2s n :: mkr (c+1) cns'
	else raise Fail "mkrow : c' < c"
      | mkr c [] = if c > nc then [] 
	else "" :: mkr (c+1) [] ;
  in mkr 1 cns end ;

fun mkfrow' nc (Cres {count = ((tot, (0, nexh) :: cns)), tv, ...}) =
  mkrow' nc cns @ [(i2s nexh), (i2s tot), (tvprint tv) ] ;

fun mkrow nc ctc = List.map (fn "" => "&nbsp;" | s => s) (mkrow' nc ctc) ;
fun mkfrow nc ctc = List.map (fn "" => "&nbsp;" | s => s) (mkfrow' nc ctc) ;

(* mkscrrow' : int -> int * results -> wseq list
  mksccsvrow : int -> int * results -> string
  make row of scrutiny sheet from one results #count entry,
  including candidates 1 to nc, then exhausted votes and total,
  mkscrrow' in html, mksccsvrow csv format *)
fun mkscrrow' nc (ct, ctc) = 
  tda "align=center" ($ (i2s ct)) :: List.map (td o $) (mkfrow nc ctc) ;

fun mksccsvrow nc (ct, ctc) = String.concat 
  (Gen.separate "," (i2s ct :: mkfrow' nc ctc)) ;

(* prsl : string list -> string -> unit
  prints sl to named file (creates intermediate string) *)
fun prsl sl file = 
  let val ostr = TextIO.openOut file ;
    fun prl l = (TextIO.output (ostr, l) ; TextIO.output1 (ostr, #"\n")) ;
    val () = List.app prl sl ;
  in TextIO.closeOut ostr end ;

(* scrcsv : info -> string -> string list
  make scrutiny sheet in csv format *)
fun scrcsv (info as Info {results, fixed = {nc, ...}, ...}) base =
  let val rows = List.map (mksccsvrow nc) (revres info) ;
    val crow = implode (Gen.replicate #"," (nc + 3)) ;
    val crows = Gen.separate crow rows ;
    val rows' = 
      List.map (fn (ct, ctc) => (i2s ct :: mkfrow' nc ctc)) (revres info) ;
  in prsl crows (base ^ ".csv") ; rows' end ;

(* scrsh : info -> wseq, make scrutiny sheet *)
fun scrsh (info as Info {results, fixed = {nc, ...}, ...}) =
  let val rows = 
      List.map (srowc "align=right" o mkscrrow' nc) (revres info) ;
    (* not using table header, some error (mine or Netscape?) 
    val hdrow = shead "align=center" (scrheads nc) ; *)
    val hdrow = srow "align=center" (scrheads nc) ;
    val tab = tablea "border" (wsc (hdrow :: rows)) ;
  in htmldoc ($ "Scrutiny Sheet") 
    ((h1 o center) ($ "Scrutiny Sheet") && tab) end ;

(* prwseq : wseq -> string -> unit
  prints wseq to named file (creates intermediate string) *)
fun prwseq wseq file = 
  let val ostr = TextIO.openOut file ;
    val () = TextIO.output (ostr, flatten wseq) ;
  in TextIO.closeOut ostr end ;

(* prscr : info -> string -> unit
  print scrutiny sheet, HTML format, into named file *)
fun prscr info basename = prwseq (scrsh info) (basename ^ ".scr.html") ;

(*
prscr info "votefile" ;
*)
(* distheads : int -> string list, make headers of distn sheet *)
fun distheads nc = "Cnt" :: 
  List.map i2s (Gen.upto (1, nc)) @ 
  ["Exh", "LbF", "Total"] ;

(* mksl : int -> int * (int * int) list -> string list
  make list of strings for top or bottom of part of row where two values *)

fun mktop (nc, nv) (tot, (0, nexh) :: cns) =
  mkrow nc cns @ [i2s nexh, i2s (nv - tot), i2s nv] ;

fun mkbot nc (tot, (0, nexh) :: cns) =
  mkrow nc cns @ [i2s nexh, i2s (~tot), ""] ;
  
(* mkdistrow : int -> int * results -> int * results -> wseq
  make row, after first, of distribution sheet from one #new and previous
  #tot entry in results *)
fun mkdistrow (nc, nv) (ct, ni) (ct', ti) = 
  let 
    (* row of table involves #new of this count and #tot of previous count *)
    val () = 
      if ct <> ct' + 1 then raise Fail "mkdistrow: ct <> ct' + 1" else () ;  
    val top = mktop (nc, nv) ti ;
    val bot = mkbot nc ni ;
    fun f (t, b) = td ($ t && br && $ b) ;
    val cells = ListPair.map f (top, bot) ;
  in tra "align=right valign=top"
    (wsc ((tda "align=center" o $ o i2s) ct :: cells) && Nl) end ;

(* mkfirst, mklast : int * int -> int * results -> wseq
  make first or last row of distribution sheet from #tot result *)
fun mkfirst (nc, nv) (ct, ti) = 
  tra "align=right" 
  (wsc ((tda "align=center" o $) "1" :: 
    List.map (td o $) (mktop (nc, nv) ti)) && Nl) ;

fun mklast (nc, nv) (ct, ti) = 
  tra "align=right" 
  (wsc ((td o $) "&nbsp;" :: List.map (td o $) (mktop (nc, nv) ti)) && Nl) ;

(* get rows from news and tots *)
(* ntr', ntr : int * int -> 
  (int * results) list -> (int * results) list -> wseq list
  using the #new and #tot entries in results, create rows of distn sheet,
  ntr' creates all except first row (but including last row);
  ntr also creates first row *)
fun ntr' ncv [] [tot] = 
    [mklast ncv tot] (* last row of dist'n sheet *)
  | ntr' ncv (new :: news) (tot :: tots) = 
    mkdistrow ncv new tot :: ntr' ncv news tots ;

fun ntr ncv ((1, _) :: news) (tots as (tot as (1, _)) :: _) =
  mkfirst ncv tot :: (* first row of dist'n sheet *)
  ntr' ncv news tots ; (* rest of rows *)

fun distsh (info as Info {results, fixed = {nc, nv, ...}, ...}) =
  let val hdrow = srow "align=center" (distheads nc) ;
    val rows = ntr (nc, nv) (getfds #new info) (getfds #tot info) ;
    val tab = tablea "border" (wsc (hdrow :: rows)) ;
  in htmldoc ($ "Distribution of Votes") 
    ((h1 o center) ($ "Distribution of Votes") && tab) end ;

(* prdist : info -> string -> unit
  print distribution sheet, HTML format, into named file *)
fun prdist info basename = prwseq (distsh info) (basename ^ ".dst.html") ;

(*
prdist info "votefile" ;
*)

end ; (* local open Msp ; infix && ; open Types ; *) 
end ; (* structure Tab *)

