
(*
app load ["Ran", "RPV", "Count"] ;
*)

(* test and examples *)

(* to create a set of votes, of random lengths, and where candidates
  have random popularities, nv - number of votes, nc = number of candidates,
  ne = number to be elected (used later);
  note - creating random votes is slow *)
val nc = 49 ; val nv = 90000 ; val ne = 7 ;
val votes' = Ran.ranpvotes nc nv ;

(* to store these in a file *)
val () = RPV.printsvotes votes' "votefile" ;

(* to read them from a file, nv = max number to be read,
  nv = ~1 means read all *)
val (nfv, votes) = RPV.readsvotes "votefile" nv ;

(* should have read what was written *)
votes = votes' ;

(* to count the votes, this will create functions
  r i gives counting results from stage i
    basically New and Total are the figures in the boxes in table II
    Count gives the figures in the boxes in the scrutiny sheet
  e i gives events recorded for stage i
  info gives entire information structure for the count
  csl is the final status - candidates "Elected" reqched a quota,
    candidates "Continuing" were left when counting stopped,
    note - as at writing this, too many can be elected all at one count,
    need to fix this 
  #events (info) gives a list of events, and counts when they occurred
  *)

open Types Count ;

local
  val info as Info (inforec as
    {candStatus : status array, candTot : int array, count : int ref,
    events : (int * elecEvent) list ref,
    fixed : {nc : int, ne : int, nq : int}, ncont : int ref, nelec, 
    results : (int * results) list ref}) = 
    prefCount (nc, ne, nv, votes) ;
  val csl' = (Count.numl 0 
    (List.map Types.statusString (Array.foldr (op ::) [] candStatus))) ;
in
  fun r i = List.filter (fn (c, _) => c = i) (!results) ;
  fun e i = List.filter (fn (c, _) => c = i) (!events) ;
  val csl = List.filter 
    (fn (0, _) => false | (_, st) => st <> "Excluded") csl' ;
  fun i () = inforec ;
end ;

val info = Info (i ()) ;

fun ev inff = 
  let val Types.Info {events, ...} = inff () ;
  in ! events end ;

fun res inff = 
  let val Types.Info {results, ...} = inff () ;
  in ! results end ;

val results = res infof ;
val events = ev infof ;


val nc = 49 ; val nv = 90000 ; val ne = 7 ;
val file = "votefile" ;
val (nfv, votes) = RPV.readsvotes file nv ;

fun commLot cnt cands = cnt mod length cands ;

local val info = Count.prefCount (nc, ne, nfv, [], commLot, votes) ;
in fun infof () = info end ;

val () = Tab.prscr (infof ()) file ;
val () = Tab.prdist (infof ()) file ;

(*
ls -l votefile.dst votefile.scr
diff votefile.dst std/votefile.dst
diff votefile.scr std/votefile.scr
*)

val nc = 20 ; val nv = 20000 ; val ne = 7 ;
val file = "votes20" ;
val (nfv, votes) = RPV.readsvotes file nv ;

fun commLot cnt cands = cnt mod length cands ;
val decd = [11, 5, 14] ;

local val info = Count.prefCount (nc, ne, nfv, decd, commLot, votes) ;
in fun infof () = info end ;

val () = Gen.printElec (infof ()) ;

val () = Tab.prscr (infof ()) file ;
val () = Tab.prdist (infof ()) file ;


(*
ls -l votes20.dst.html votes20.scr.html
diff votes20.dst.html std/votes20.dst
diff votes20.scr.html std/votes20.scr
*)

fun findint (nc, ne, nv, decd, commLot, votes) = 
  let val nvotes = List.drop (votes, 100) ;
    val nnv = nv - 100 ;
    val info as Types.Info {events = ref events, ...} =
      Count.prefCount (nc, ne, nnv, decd, commLot, nvotes) ;
    val isInter = List.exists (Types.inter o #2) events ;
  in if isInter then (nnv, nvotes, fn () => info) 
    else findint (nc, ne, nnv, decd, commLot, nvotes) end ; 
    
val nnv = nv ; val nvotes = votes ;

fun pr (nnv, nvotes, infof) = 
  let 
    val () = Tab.prscr (infof ()) ("nv" ^ Int.toString nnv) ;
    val () = Tab.prdist (infof ()) ("nv" ^ Int.toString nnv) ;
  in () end ;

val (nnv, nvotes, infof) = findint (nc, ne, nnv, decd, commLot, nvotes) ;
val inters = List.filter (Types.inter o #2) (ev infof) ;
val () = pr (nnv, nvotes, infof) ;


(** for testing small numbers of votes, 
  this routine repeatedly removes 1 vote and retests **)
fun doall (nc, ne, nv, decd, commLot, votes) = 
  let 
    val info = Count.prefCount (nc, ne, nv, decd, commLot, votes) ;
  in case (* TextIO.inputLine TextIO.stdIn of "\n" *) () of () => 
	doall (nc, ne, nv - 1, decd, commLot, tl votes) 
      | _ => (nv, votes) end 
  handle exn => (print (exnMessage exn) ; (nv, votes)) ;
    
val (nnv, nvotes) = doall (nc, ne, nv, decd, commLot, votes) ; 

val info = Count.prefCount (nc, ne, nnv, decd, commLot, nvotes) ;

